/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: ClientHull.hpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#pragma once

#include "Hull.hpp"
#include "common.hpp"
#include "TCPConnection.hpp"
#include <array>


namespace EPS
{
	/// Client hull. @see EPS::Hull @see EPS::ServerHull
	class ClientHull final : public Hull
	{
		public:

			/// Constructor.
			ClientHull(int argc, char *argv[]);

			/// Destructor.
			virtual ~ClientHull(void);

			/// Run EPS.
			virtual void run_second_stage(void);

			/// Find the server and start one or more new TCP beams.
			virtual void async_find_and_connect_to_server(void);

			/// Callback when a connection has failed and needs to be cleaned up.
			virtual void connection_failed(const EPS::TConnPtr &connection);

			/// Called when a communication command is received.
			virtual void processComm(const EPS::TConnPtr &connection, EPS::NetBuffer *nb, const EPS::CommHeader1 *header, const EPS::ECommType type);

			/// Remove all of our entries from the hosts file.
			std::string clean_hosts_file(std::string hosts = "");

			/// Add or remove the necessary entries into the hosts file.
			void update_hosts_file(void);

			/// Asynchronous read from the TUN/TAP interface.
			void async_read_tun_tap(void);

			/// Called once a packet has been read from the TUN/TAP interface.
			void process_tun_tap_packet(EPS::NetBuffer *nb);

			// members

			std::string					device_name;	///< TUN/TAP device name (such as "eps0")
			FILE_HANDLE_DESCRIPTOR		device_fd;		///< raw file handle or descriptor for TUN/TAP device
			ASIO_STREAM					tun_tap_stream;	///< TUN/TAP stream handle for ASIO read/write
	};
}
