/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Cipher.hpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#pragma once

#include "common.hpp"
#include "NetBuffer.hpp"
#include <openssl/evp.h>


/// Definition needed when calling OpenSSL functions such as EVP_CipherInit_ex() @{
#define ENC_MODE_UNCHANGED	-1
#define ENC_MODE_DECRYPT	0
#define ENC_MODE_ENCRYPT	1
/// @}


namespace EPS
{
	/// Cipher takes care of calling OpenSSL to encrypt and decrypt @ref NetBuffer objects.
	namespace Cipher
	{
		/// Reset OpenSSL.
		void reset(void);

		/// Initialize OpenSSL ciphers.
		void initialize(EPS::StrMap &configuration);

		/// Get and parse the key and iv from configuration.
		void get_cipher_key_and_iv(EPS::StrMap &configuration);

		/// Encrypt a network buffer.  Return a pointer to the newly-encrypted buffer.  The original buffer remains untouched.
		EPS::NetBuffer *encrypt(EPS::NetBuffer &buffer);

		/// Decrypt a network buffer.  Return a pointer to the newly-decrypted buffer.  The original buffer remains untouched.
		EPS::NetBuffer *decrypt(EPS::NetBuffer &buffer);

		/// Decrypt only the first few bytes ("header") of a network buffer.  Return a pointer to the newly-decrypted buffer.  The original buffer remains untouched.
		EPS::NetBuffer *decrypt_header(EPS::NetBuffer &buffer);

		extern std::array<uint8_t, 32>	key;					///< OpenSSL:  256-bit encryption/decryption key
		extern std::array<uint8_t, 16>	iv;						///< OpenSSL:  128-bit encryption/decryption iv
		extern const EVP_CIPHER		*	cipher;					///< OpenSSL:  cipher to use for encryption/decryption
		extern EVP_CIPHER_CTX		*	decrypt_ctx;			///< OpenSSL:  ctx needed for decryption
		extern EVP_CIPHER_CTX		*	encrypt_ctx;			///< OpenSSL:  ctx needed for encryption
		extern size_t					size_of_keys_in_bytes;	///< OpenSSL:  size of the cipher key
		extern size_t					size_of_iv_in_bytes;	///< OpenSSL:  size of the cipher initialization vector
		extern std::string				name;					///< OpenSSL:  text name of the cipher in use
		extern std::string				hash;					///< nonsense text string to help validate encryption key and iv
	}
}
