/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Statistics.cpp 2182 2017-03-26 08:30:05Z stephane $
 * See license.txt.
 */

#include "Statistics.hpp"
#include "logging.hpp"
#include "Hull.hpp"
#include <sstream>
#include <ctime>


EPS::Statistics::Statistics(void)
{
	this->clear();

	return;
}


void EPS::Statistics::clear(void)
{
	memset(this, '\0', sizeof(EPS::Statistics));

	return;
}


EPS::Statistics &EPS::Statistics::operator+=(const Statistics &rhs)
{
	this->server_packets_forwarded	+= rhs.server_packets_forwarded;
	this->server_bytes_forwarded	+= rhs.server_bytes_forwarded;
	this->server_packets_received	+= rhs.server_packets_received;
	this->server_bytes_received		+= rhs.server_bytes_received;
	this->server_packets_sent		+= rhs.server_packets_sent;
	this->server_bytes_sent			+= rhs.server_bytes_sent;

	this->packets_dropped			+= rhs.packets_dropped;
	this->bytes_dropped				+= rhs.bytes_dropped;

	this->direct_packets_received	+= rhs.direct_packets_received;
	this->direct_bytes_received		+= rhs.direct_bytes_received;
	this->direct_packets_sent		+= rhs.direct_packets_sent;
	this->direct_bytes_sent			+= rhs.direct_bytes_sent;

	this->tap_packets_received		+= rhs.tap_packets_received;
	this->tap_bytes_received		+= rhs.tap_bytes_received;
	this->tap_packets_sent			+= rhs.tap_packets_sent;
	this->tap_bytes_sent			+= rhs.tap_bytes_sent;

	this->connections_established	+= rhs.connections_established;
	this->connections_dropped		+= rhs.connections_dropped;

	return *this;
}


std::string EPS::Statistics::to_string(void) const
{
	std::stringstream ss;

	// only include non-zero stats

	if (server_packets_forwarded	) ss << std::endl << "\t" << server_packets_forwarded	<< " server packets forwarded";
	if (server_bytes_forwarded		) ss << std::endl << "\t" << server_bytes_forwarded		<< " server bytes forwarded ("	<< format_bytes(server_bytes_forwarded)	<< ")";
	if (server_packets_received		) ss << std::endl << "\t" << server_packets_received	<< " server packets received";
	if (server_bytes_received		) ss << std::endl << "\t" << server_bytes_received		<< " server bytes received ("	<< format_bytes(server_bytes_received)	<< ")";
	if (server_packets_sent			) ss << std::endl << "\t" << server_packets_sent		<< " server packets sent";
	if (server_bytes_sent			) ss << std::endl << "\t" << server_bytes_sent			<< " server bytes sent ("		<< format_bytes(server_bytes_sent)		<< ")";

	if (packets_dropped				) ss << std::endl << "\t" << packets_dropped			<< " packets dropped";
	if (bytes_dropped				) ss << std::endl << "\t" << bytes_dropped				<< " bytes dropped ("			<< format_bytes(bytes_dropped)			<< ")";

	if (direct_packets_received		) ss << std::endl << "\t" << direct_packets_received	<< " direct packets received";
	if (direct_bytes_received		) ss << std::endl << "\t" << direct_bytes_received		<< " direct bytes received ("	<< format_bytes(direct_bytes_received)	<< ")";
	if (direct_packets_sent			) ss << std::endl << "\t" << direct_packets_sent		<< " direct packets sent";
	if (direct_bytes_sent			) ss << std::endl << "\t" << direct_bytes_sent			<< " direct bytes sent ("		<< format_bytes(direct_bytes_sent)		<< ")";

	if (tap_packets_received		) ss << std::endl << "\t" << tap_packets_received		<< " tap packet received";
	if (tap_bytes_received			) ss << std::endl << "\t" << tap_bytes_received			<< " tap bytes received ("		<< format_bytes(tap_bytes_received)		<< ")";
	if (tap_packets_sent			) ss << std::endl << "\t" << tap_packets_sent			<< " tap packets sent";
	if (tap_bytes_sent				) ss << std::endl << "\t" << tap_bytes_sent				<< " tap bytes sent ("			<< format_bytes(tap_bytes_sent)			<< ")";

	if (connections_established		) ss << std::endl << "\t" << connections_established	<< " connections established";
	if (connections_dropped			) ss << std::endl << "\t" << connections_dropped		<< " connections dropped";

	if (server_bps_in	[0] != '\0'	) ss << std::endl << "\tserver bps received: "	<< server_bps_in;
	if (server_bps_out	[0] != '\0'	) ss << std::endl << "\tserver bps sent: "		<< server_bps_out;
	if (server_bps_fwd	[0] != '\0'	) ss << std::endl << "\tserver bps forwarded: "	<< server_bps_fwd;
	if (direct_bps_in	[0] != '\0'	) ss << std::endl << "\tdirect bps received: "	<< direct_bps_in;
	if (direct_bps_out	[0] != '\0' ) ss << std::endl << "\tdirect bps sent: "		<< direct_bps_out;
	if (tap_bps_in		[0] != '\0' ) ss << std::endl << "\ttap bps received: "		<< tap_bps_in;
	if (tap_bps_out		[0] != '\0' ) ss << std::endl << "\ttap bps sent: "			<< tap_bps_out;

	return ss.str();
}


bool EPS::Statistics::empty(void) const
{
	return	server_packets_forwarded	== 0 &&
			server_bytes_forwarded		== 0 &&
			server_packets_received		== 0 &&
			server_bytes_received		== 0 &&
			server_packets_sent			== 0 &&
			server_bytes_sent			== 0 &&
			packets_dropped				== 0 &&
			bytes_dropped				== 0 &&
			direct_packets_received		== 0 &&
			direct_bytes_received		== 0 &&
			direct_packets_sent			== 0 &&
			direct_bytes_sent			== 0 &&
			tap_packets_received		== 0 &&
			tap_bytes_received			== 0 &&
			tap_packets_sent			== 0 &&
			tap_bytes_sent				== 0 &&
			connections_established		== 0 &&
			connections_dropped			== 0 ;
}


void EPS::Statistics::export_scorecard(std::ofstream &fs, const std::string &prefix) const
{
	fs	<< prefix << "server_packets_forwarded="		<<				server_packets_forwarded				<< std::endl
		<< prefix << "server_bytes_forwarded="			<<				server_bytes_forwarded					<< std::endl
		<< prefix << "server_bytes_forwarded_text=\""	<< format_bytes(server_bytes_forwarded		)	<< "\""	<< std::endl
		<< prefix << "server_packets_received="			<<				server_packets_received					<< std::endl
		<< prefix << "server_bytes_received="			<<				server_bytes_received					<< std::endl
		<< prefix << "server_bytes_received_text=\""	<< format_bytes(server_bytes_received		)	<< "\""	<< std::endl
		<< prefix << "server_packets_sent="				<<				server_packets_sent						<< std::endl
		<< prefix << "server_bytes_sent="				<<				server_bytes_sent						<< std::endl
		<< prefix << "server_bytes_sent_text=\""		<< format_bytes(server_bytes_sent			)	<< "\""	<< std::endl

		<< prefix << "packets_dropped="					<<				packets_dropped							<< std::endl
		<< prefix << "bytes_dropped="					<<				bytes_dropped							<< std::endl
		<< prefix << "bytes_dropped_text=\""			<< format_bytes(bytes_dropped				)	<< "\""	<< std::endl

		<< prefix << "direct_packets_received="			<<				direct_packets_received					<< std::endl
		<< prefix << "direct_bytes_received="			<<				direct_bytes_received					<< std::endl
		<< prefix << "direct_bytes_received_text=\""	<< format_bytes(direct_bytes_received		)	<< "\""	<< std::endl
		<< prefix << "direct_packets_sent="				<<				direct_packets_sent						<< std::endl
		<< prefix << "direct_bytes_sent="				<<				direct_bytes_sent						<< std::endl
		<< prefix << "direct_bytes_sent_text=\""		<< format_bytes(direct_bytes_sent			)	<< "\""	<< std::endl

		<< prefix << "tap_packets_received="			<<				tap_packets_received					<< std::endl
		<< prefix << "tap_bytes_received="				<<				tap_bytes_received						<< std::endl
		<< prefix << "tap_bytes_received_text=\""		<< format_bytes(tap_bytes_received			)	<< "\""	<< std::endl
		<< prefix << "tap_packets_sent="				<<				tap_packets_sent						<< std::endl
		<< prefix << "tap_bytes_sent="					<<				tap_bytes_sent							<< std::endl
		<< prefix << "tap_bytes_sent_text=\""			<< format_bytes(tap_bytes_sent				)	<< "\""	<< std::endl

		<< prefix << "connections_established="			<<				connections_established					<< std::endl
		<< prefix << "connections_dropped="				<<				connections_dropped						<< std::endl;

	// are the BPS values stored in this stats record?
	if (	server_bps_in	[0] != '\0' ||
			server_bps_out	[0] != '\0' ||
			direct_bps_in	[0] != '\0' ||
			direct_bps_out	[0] != '\0' ||
			tap_bps_in		[0] != '\0' ||
			tap_bps_out		[0] != '\0' )
	{
		fs	<< std::endl
			<< "server_bps_in="		"\"" << server_bps_in	<< "\"" << std::endl
			<< "server_bps_out="	"\"" << server_bps_out	<< "\"" << std::endl
			<< "server_bps_fwd="	"\"" << server_bps_fwd	<< "\"" << std::endl
			<< "direct_bps_in="		"\"" << direct_bps_in	<< "\"" << std::endl
			<< "direct_bps_out="	"\"" << direct_bps_out	<< "\"" << std::endl
			<< "tap_bps_in="		"\"" << tap_bps_in		<< "\"" << std::endl
			<< "tap_bps_out="		"\"" << tap_bps_out		<< "\"" << std::endl
			<< std::endl;
	}

	return;
}


EPS::Stats::Stats(void) :
	log_stats			(false),
	scorecard_enabled	(false)
{
	return;
}


std::string EPS::Stats::to_string(void) const
{
	std::stringstream ss;
	ss	<< "60-second stats:"	<< this->minute_stats	.to_string() << std::endl
		<< "total stats:"		<< this->total_stats	.to_string();

	return ss.str();
}


bool EPS::Stats::empty(void) const
{
	return this->total_stats.empty() && this->minute_stats.empty();
}


void EPS::Stats::roll(void)
{
	// if there are no new stats, we can skip some things
	static bool	show_stats_anyway		= true;
	const bool	new_stats_are_empty		= this->minute_stats.empty();
	const bool	new_stats_are_available	= ! new_stats_are_empty;
	const bool	record_statistics		= new_stats_are_available || show_stats_anyway;

	if (new_stats_are_available)
	{
		// copy the stats into the totals
		this->total_stats += this->minute_stats;
		show_stats_anyway = true;
	}
	if (new_stats_are_empty)
	{
		// next time through, we don't want to force showing empty stats
		show_stats_anyway = false;
	}

	const uint64_t server_bytes_in		= this->minute_stats.server_bytes_received;
	const uint64_t server_bytes_out		= this->minute_stats.server_bytes_sent;
	const uint64_t server_bytes_fwd		= this->minute_stats.server_bytes_forwarded;
	const uint64_t direct_bytes_in		= this->minute_stats.direct_bytes_received;
	const uint64_t direct_bytes_out		= this->minute_stats.direct_bytes_sent;
	const uint64_t tap_bytes_in			= this->minute_stats.tap_bytes_received;
	const uint64_t tap_bytes_out		= this->minute_stats.tap_bytes_sent;

	strncpy(this->total_stats.server_bps_in	, EPS::format_bps(server_bytes_in	, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.server_bps_out, EPS::format_bps(server_bytes_out	, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.server_bps_fwd, EPS::format_bps(server_bytes_fwd	, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.direct_bps_in	, EPS::format_bps(direct_bytes_in	, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.direct_bps_out, EPS::format_bps(direct_bytes_out	, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.tap_bps_in	, EPS::format_bps(tap_bytes_in		, 60).c_str(), EPS::bps_string_size);
	strncpy(this->total_stats.tap_bps_out	, EPS::format_bps(tap_bytes_out		, 60).c_str(), EPS::bps_string_size);

	// *********************************
	// record all stats to the scorecard
	// *********************************
	if (this->scorecard_enabled && not this->scorecard_filename.empty())
	{
		const EPS::Hull &hull = EPS::Hull::get();

		const time_t now = std::time(nullptr);

		static const time_t			startup_time	= hull.start_time;
		static const std::string	mode			= hull.mode_string;
		static const pid_t			pid				= getpid();

		std::string tmp = this->scorecard_filename + "_tmp";
		std::ofstream fs(tmp);
		fs	<< "#!/bin/bash"																		<< std::endl
			<< ""																					<< std::endl
			<< "timestamp="				<< "\""<< EPS::get_utc_timestamp(now) << "\""				<< std::endl
			<< "timestamp_in_seconds="	<< now														<< std::endl
			<< "eps_uptime_in_seconds="	<< now - startup_time 										<< std::endl
			<< "eps_uptime_duration="	<< "\"" << EPS::get_approximate_time(startup_time) << "\""	<< std::endl
			<< "eps_mode="				<< mode														<< std::endl
			<< "eps_ver="				<< EPS_VERSION												<< std::endl
			<< "eps_pid="				<< pid														<< std::endl
			<< ""																					<< std::endl;

		this->minute_stats	.export_scorecard(fs, "minute_"	);
		fs << std::endl;
		this->total_stats	.export_scorecard(fs, "total_"	);

		// ********* FOR DEBUG PURPOSE ONLY *********
		// eventually this block of code needs to be removed
		if (not hull.tracker.empty())
		{
			// add 1 row per client for debug reasons
			fs << "# the following peer information is included for debug reasons, and will be removed in a future release" << std::endl;

			for (const auto iter : hull.tracker.get_all_entries())
			{
				const EPS::TrackerData &data = iter.second;

				fs << "# " << data.as_debug() << std::endl;
			}
		}
		// end of debug code to be removed
		// ********* FOR DEBUG PURPOSE ONLY *********

		fs << std::endl;
		fs.close();
		rename(tmp.c_str(), this->scorecard_filename.c_str());
	}

	// *******************************************
	// record all stats to the normal eps.log file
	// *******************************************
	if (record_statistics)
	{
		if (this->log_stats)	// do we want to log the "long" version of the stats?
		{
			LOG_INFO << "EPS statistics:" << std::endl << this->to_string();
		}

		if (new_stats_are_empty)
		{
			LOG_INFO << "rate: " << EPS::format_bps(0, 60) << " (no eps conduit network traffic recorded)";
		}
		else
		{
			std::stringstream ss;
			if ( server_bytes_in	) ss << " server_in="	<< this->total_stats.server_bps_in;
			if ( server_bytes_out	) ss << " server_out="	<< this->total_stats.server_bps_out;
			if ( server_bytes_fwd	) ss << " server_fwd="	<< this->total_stats.server_bps_fwd;
			if ( direct_bytes_in	) ss << " direct_in="	<< this->total_stats.direct_bps_in;
			if ( direct_bytes_out	) ss << " direct_out="	<< this->total_stats.direct_bps_out;
			if ( tap_bytes_in		) ss << " tap_in="		<< this->total_stats.tap_bps_in;
			if ( tap_bytes_out		) ss << " tap_out="		<< this->total_stats.tap_bps_out;

			LOG_INFO << "rate:" << ss.str();
		}
	}

	if (new_stats_are_available)
	{
		// wait until we've printed out all the statistics before we clear the 60-second stats
		this->minute_stats.clear();
	}

	return;
}
